;; ediprolog -- Emacs does Interactive Prolog
;; Written Nov. 28th, 2006 by Markus Triska (triska@gmx.at)
;; Public domain code.

;; These definitions let you transparently interact with SWI Prolog in
;; all buffers.  You can load and syntax-check Prolog programs and
;; execute queries with minimal exposure to the toplevel.  Queries
;; start with `?-' or `:-', possibly preceded by `%' and whitespace.

;; Copy this file to your home directory and add to your .emacs:

;;     (load "~/ediprolog")
;;     (global-set-key [f10] 'ediprolog-dwim)

;; Restart Emacs and press F10 in a Prolog program to load it (point
;; is moved to first error, if there are any).  Press F10 on a query
;; in the program to execute it. Query results are inserted into the
;; buffer, and you interact with SWI Prolog as on a terminal.

;; Tested with SWI Prolog 5.6.24 and Emacs 21.4 and 22.0.


(defvar ediprolog-prefix "%@% "
  "String to prepend when inserting output from the Prolog
process into the buffer.")

(defvar ediprolog-command-line-parameters "-q"
  "Command line arguments to pass to the Prolog executable. For
SWI Prolog, `-q' must be included.")

(defvar ediprolog-prolog-executable "pl"
  "Program file name if the Prolog executable.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ediprolog-more-solutions nil)

(defvar ediprolog-process nil)

(defvar ediprolog-consult-error nil)

(defvar ediprolog-version "0.8c")

(defvar ediprolog-read-term nil)

(defmacro ediprolog-wait-for-prompt-after (form)
  "Executes FORM and waits for prompt."
  `(progn
     (setq ediprolog-seen-prompt nil)
     ,form
     (set-process-filter ediprolog-process 'ediprolog-wait-for-prompt-filter)
     (while (not ediprolog-seen-prompt)
       (accept-process-output ediprolog-process))))

(defun ediprolog-run-prolog ()
  (ediprolog-wait-for-prompt-after
   (setq ediprolog-process
	 (start-process "ediprolog" nil ediprolog-prolog-executable
			ediprolog-command-line-parameters))))

(defun ediprolog-wait-for-response ()
  (accept-process-output ediprolog-process))

(defun ediprolog-wait-for-prompt-filter (proc string)
  (if (member "?- " (split-string string "\n"))
      (setq ediprolog-seen-prompt t)))

(defun ediprolog-fresh-process ()
  (if (and (processp ediprolog-process)
	   (eq (process-status ediprolog-process) 'run))
      (kill-process ediprolog-process))
  (ediprolog-run-prolog))

(defun ediprolog-ensure-process ()
  (if (or (not (processp ediprolog-process))
	  (not (eq (process-status ediprolog-process) 'run)))
      (ediprolog-run-prolog)))


(defun ediprolog-dwim (&optional arg)
  "Load current buffer into Prolog or post query (Do What I Mean).
If invoked on a line starting with `:-' or `?-', possibly
preceded by `%' and/or whitespace, send the query to the Prolog
process and start interaction. Otherwise, call
`ediprolog-consult'.

If invoked with a prefix argument, only remove all previous
interactions from the buffer. With prefix argument zero, also
kill the Prolog process if it is running."
  (interactive "P")
  (cond (arg
	 (ediprolog-remove-output)
	 (if (and (eq arg 0) (processp ediprolog-process)
		  (eq (process-status ediprolog-process) 'run))
	     (kill-process ediprolog-process)))
	(t
	 (let ((line (thing-at-point 'line)))
	   (if (string-match "^\\s *%*\\s *[:?]-\\s *\\(.*\\)$" line)
	       (ediprolog-interact (format "%s\n" (match-string 1 line)))
	     (ediprolog-consult))))))

(defun ediprolog-interact (query)
  (ediprolog-ensure-process)
  (end-of-line)
  ; set verbose flag to get access to time/1 information
  (ediprolog-wait-for-prompt-after
   (process-send-string ediprolog-process "set_prolog_flag(verbose,normal).\n"))
  (set-process-filter ediprolog-process 'ediprolog-interact-filter)
  (setq ediprolog-more-solutions t)
  (setq ediprolog-read-term nil)
  (process-send-string ediprolog-process query)
  (ediprolog-wait-for-response)
  (sit-for 0.1)
  (while ediprolog-more-solutions
    (let (string char)
      (if ediprolog-read-term
	  (setq string (concat (read-string "Input: ") "\n"))
	(while (not (char-valid-p char))
	  (setq char (read-event))
	  (if (char-valid-p char)
	      (progn
		(insert char)
		(setq string (char-to-string char)))
	    (message "Press h for help."))))
      (setq ediprolog-read-term nil)
      (if ediprolog-more-solutions
	  (process-send-string ediprolog-process string))
      (when ediprolog-more-solutions
	(ediprolog-wait-for-response)
	(sit-for 0.1)	    ; give prompt ("?- ") a chance
	))))


(defun ediprolog-remove-output ()
  "Remove all lines starting with ediprolog-prefix from buffer."
  (interactive)
  (save-excursion
    (goto-char (point-min))
    (while (re-search-forward (concat "^" ediprolog-prefix ".*\n?") nil t)
      (replace-match "" nil nil))))

(defun ediprolog-consult ()
  "Loads buffer into a fresh Prolog process.
First, all previous interactions are removed from the
buffer. Then, buffer contents are sent line by line to a fresh
Prolog process. Any output from the Prolog process not equal
`|: ', `Yes' or `?- ' is interpreted as error, resulting in not
proceeding any further, displaying the output, and moving point
to a line close to where the error occurred. Therefore,
`ediprolog-command-line-parameters' must include `-q'.

In transient mark mode, the function operates on the region if it
is active."
  (interactive)
  (ediprolog-fresh-process)
  (set-process-filter ediprolog-process 'ediprolog-consult-filter)
  (process-send-string ediprolog-process "[user].\n")
  (ediprolog-wait-for-response)
  (setq ediprolog-consult-error nil)
  (let ((ediprolog-error-line 1))
    (save-excursion
      (save-restriction
	(if (and transient-mark-mode mark-active)
	    (narrow-to-region (region-beginning) (region-end)))
	(ediprolog-remove-output)
	(goto-char (point-min))
	(while (and (not (eobp)) (not ediprolog-consult-error))
	  (let ((line (thing-at-point 'line)))
	    (process-send-string ediprolog-process line)
	    (unless (string-match "\n$" line)
	      (process-send-string ediprolog-process "\n")))
	  (ediprolog-wait-for-response)
	  (forward-line)
	  (setq ediprolog-error-line (1+ ediprolog-error-line)))))
    (if ediprolog-consult-error
    (progn
      (if (> ediprolog-error-line 1)
	  (setq ediprolog-error-line (1- ediprolog-error-line)))
      (if (and transient-mark-mode mark-active)
	  (princ (format "\nProblem near line %s of region." ediprolog-error-line))
	(goto-line ediprolog-error-line))
      (sit-for 1))
    (message (concat (if (and transient-mark-mode mark-active)
			 "Region" "Buffer") " consulted.")))
    (ediprolog-wait-for-prompt-after (process-send-eof ediprolog-process))))

(defun ediprolog-running ()
  (and (processp ediprolog-process)
       (eq (process-status ediprolog-process) 'run)))

(defun ediprolog-consult-filter (proc string)
  (let ((lines (split-string string "\n")))
    (mapc (lambda (el)
	    (setq lines (delete el lines))) '("" "|: " "Yes" "?- "))
    (when lines
      (mapc 'princ lines)
      (setq ediprolog-consult-error t))))

(defun ediprolog-interact-filter (proc string)
  (let ((lines (split-string string "\n")))
    (if (member "?- " lines) (setq ediprolog-more-solutions nil))
    (if (member "|: " lines) (setq ediprolog-read-term t))
;;     (print lines)
    (mapc (lambda (el)
	    (setq lines (delete el lines))) '("" "[print]" "[write]" ";"
					      "?- " " " "|: "))
    (mapc (lambda (el)
	    (end-of-line)
	    (insert (concat "\n" ediprolog-prefix el))) lines)))

(provide 'ediprolog)
